import pydantic
import param
from panel.io import init_doc, state

from panel.layout import Column, Panel, Card, WidgetBox

from panel.pane import PaneBase

from .widgets import PydanticModelEditor, PydanticModelEditorCard
from .dispatchers import get_widget

from typing import (
    Any,
    Optional,
)


from bokeh.document import Document
from bokeh.model import Model
from pyviz_comms import Comm


class Pydantic(PaneBase):
    """pydantic Pane so that calling pn.panel(model)
    will work. This pane is not expected to be used directly.
    """

    default_layout = param.ClassSelector(
        default=WidgetBox, class_=Panel, is_instance=False
    )

    
    object = param.Parameter(default=None)

    def __init__(self, object=None, **params):

        pane_params = {
            name: params[name] for name in Pydantic.param.params() if name in params
        }

        super().__init__(object, **pane_params)


        if isinstance(object, pydantic.BaseModel):
            self.widget = get_widget(object, None, class_=object.__class__, **params)
            self.object = object

        elif issubclass(object, pydantic.BaseModel):
            self.widget = get_widget.invoke(object, None)(None, None, class_=object, **params)
            self.widget.link(self, value="object")
        else:
            raise

        self.layout = self.default_layout(self.widget)

    def _get_model(
        self,
        doc: Document,
        root: Optional[Model] = None,
        parent: Optional[Model] = None,
        comm: Optional[Comm] = None,
    ) -> Model:
        model = self.layout._get_model(doc, root, parent, comm)
        self._models[root.ref["id"]] = (model, parent)
        return model

    def _cleanup(self, root: Optional[Model] = None) -> None:
        self.layout._cleanup(root)
        super()._cleanup(root)

    @classmethod
    def applies(cls, obj: Any, **params) -> Optional[bool]:
        if isinstance(obj, pydantic.BaseModel):
            return True
            
        if issubclass(obj, pydantic.BaseModel):
            return True
        return False

    def select(self, selector=None):
        """
        Iterates over the Viewable and any potential children in the
        applying the Selector.

        Arguments
        ---------
        selector: type or callable or None
          The selector allows selecting a subset of Viewables by
          declaring a type or callable function to filter by.

        Returns
        -------
        viewables: list(Viewable)
        """
        return super().select(selector) + self.layout.select(selector)

    def get_root(
        self,
        doc: Optional[Document] = None,
        comm: Optional[Comm] = None,
        preprocess: bool = True,
    ) -> Model:
        """
        Returns the root model and applies pre-processing hooks

        Arguments
        ---------
        doc: bokeh.Document
          Bokeh document the bokeh model will be attached to.
        comm: pyviz_comms.Comm
          Optional pyviz_comms when working in notebook
        preprocess: boolean (default=True)
          Whether to run preprocessing hooks

        Returns
        -------
        Returns the bokeh model corresponding to this panel object
        """
        doc = init_doc(doc)
        root = self.layout.get_root(doc, comm, preprocess)
        ref = root.ref["id"]
        self._models[ref] = (root, None)
        state._views[ref] = (self, root, doc, comm)
        return root



import datetime
from typing import Dict, List, Any

try:
    from typing import _LiteralGenericAlias
except ImportError:
    _LiteralGenericAlias = None

from plum import dispatch
from numbers import Integral, Number


from panel.widgets import (
    LiteralInput,
    IntInput,
    NumberInput,
    DatetimePicker,
    Checkbox,
    TextInput,
    TextAreaInput,
    ArrayInput,
    Select,
)


ListInput = type("ListInput", (LiteralInput,), {"type": list})
DictInput = type("DictInput", (LiteralInput,), {"type": dict})
TupleInput = type("TupleInput", (LiteralInput,), {"type": tuple})

def clean_kwargs(obj, kwargs):
    return { k:v for k,v in kwargs.items()
             if k in obj.param.params()}

@dispatch
def get_widget(value: Any, field: Any, **kwargs):
    """Fallback function when a more specific
    function was not registered.
    """

    if field is not None and type(field.outer_type_) == _LiteralGenericAlias:
        options = list(field.outer_type_.__args__)
        if value not in options:
            value = options[0]
        options = kwargs.pop('options', options)
        kwargs = clean_kwargs(Select, kwargs)

        return Select(value=value, options=options, **kwargs)

    kwargs = clean_kwargs(LiteralInput, kwargs)
    return LiteralInput(value=value, **kwargs)


@dispatch
def get_widget(value: Integral, field: Any, **kwargs):
    start = None
    end = None
    if field is not None:
        if type(field.outer_type_) == _LiteralGenericAlias:
            options = list(field.outer_type_.__args__)
            if value not in options:
                value = options[0]
            options = kwargs.pop('options', options)
            kwargs = clean_kwargs(Select, kwargs)
            return Select(value=value, options=options, **kwargs)

        start = getattr(field.field_info, "gt", None)
        if start is not None:
            start += 1
        else:
            start = getattr(field.field_info, "ge")

        end = getattr(field.field_info, "lt", None)
        if end is not None:
            end -= 1
        else:
            end = getattr(field.field_info, "le", None)
    kwargs = clean_kwargs(IntInput, kwargs)
    return IntInput(value=value, start=start, end=end, **kwargs)


@dispatch
def get_widget(value: Number, field: Any, **kwargs):
    start = None
    end = None
    if field is not None:
        if type(field.outer_type_) == _LiteralGenericAlias:
            options = list(field.outer_type_.__args__)
            if value not in options:
                value = options[0]
            options = kwargs.pop('options', options)
            kwargs = clean_kwargs(Select, kwargs)
            return Select(value=value, options=options, **kwargs)

        start = getattr(field.field_info, "gt", None)
        end = getattr(field.field_info, "lt", None)
    kwargs = clean_kwargs(NumberInput, kwargs)
    return NumberInput(value=value, start=start, end=end, **kwargs)


@dispatch
def get_widget(value: bool, field: Any, **kwargs):
    if value is None:
        value = False
    kwargs = clean_kwargs(Checkbox, kwargs)
    return Checkbox(value=value, **kwargs)


@dispatch
def get_widget(value: str, field: Any, **kwargs):
    min_length = kwargs.pop('min_length', None)
    max_length = kwargs.pop('max_length', 100)

    if field is not None:
        if type(field.outer_type_) == _LiteralGenericAlias:
            options = list(field.outer_type_.__args__)
            if value not in options:
                value = options[0]
            options = kwargs.pop('options', options)
            kwargs = clean_kwargs(Select, kwargs)
            return Select(value=value, options=options, **kwargs)
        max_length = field.field_info.max_length
        min_length = field.field_info.min_length
    
    kwargs['min_length']  = min_length

    if max_length is None:
        kwargs = clean_kwargs(TextAreaInput, kwargs)
        return TextAreaInput(value=value, **kwargs)

    elif max_length < 100:
        kwargs = clean_kwargs(TextInput, kwargs)
        return TextInput(value=value, max_length=max_length, **kwargs)

    kwargs = clean_kwargs(TextAreaInput, kwargs)
    return TextAreaInput(value=value, max_length=max_length, **kwargs)


@dispatch
def get_widget(value: List, field: Any, **kwargs):
    kwargs = clean_kwargs(ListInput, kwargs)
    return ListInput(value=value, **kwargs)


@dispatch
def get_widget(value: Dict, field: Any, **kwargs):
    kwargs = clean_kwargs(DictInput, kwargs)
    return DictInput(value=value, **kwargs)


@dispatch
def get_widget(value: tuple, field: Any, **kwargs):
    kwargs = clean_kwargs(TupleInput, kwargs)
    return TupleInput(value=value, **kwargs)


@dispatch
def get_widget(value: datetime.datetime, field: Any, **kwargs):
    kwargs = clean_kwargs(DatetimePicker, kwargs)
    return DatetimePicker(value=value, **kwargs)


try:
    import numpy

    @dispatch
    def get_widget(value: numpy.ndarray, field: Any, **kwargs):
        kwargs = clean_kwargs(ArrayInput, kwargs)
        return ArrayInput(value=value, **kwargs)

except ImportError:
    pass


@dispatch
def json_serializable(value: Any):
    return value


@dispatch
def json_serializable(value: list):
    return [json_serializable(v) for v in value]


@dispatch
def json_serializable(value: tuple):
    return tuple(json_serializable(v) for v in value)


@dispatch
def json_serializable(value: dict):
    return {json_serializable(k): json_serializable(v) for k, v in value.items()}


try:
    import pandas

    @dispatch
    def json_serializable(value: pandas.Interval):
        return (value.left, value.right)

except ImportError:
    pass



#!/usr/bin/env python
"""Tests for `pydantic_panel` package."""
# pylint: disable=redefined-outer-name

import pydantic_panel
import pytest
import panel as pn
from pydantic import BaseModel


class SomeModel(BaseModel):
    regular_string: str = "string"
    regular_int: int = 42
    regular_float: float = 0.999

alt_data = dict(
    regular_string = "string2",
    regular_int = 666,
    regular_float = 0.111,
)

def test_panel_model_class():
    w = pn.panel(SomeModel)
    assert isinstance(w, pydantic_panel.PydanticModelEditor)
    assert w.value == SomeModel()


def test_panel_model_instance():
    w = pn.panel(SomeModel())
    assert isinstance(w, pydantic_panel.PydanticModelEditor)
    assert w.value == SomeModel()


def test_set_data():
    m = SomeModel()
    w = pn.panel(m)
    for k,v in alt_data.items():
        w._widgets[k].value = v
        assert getattr(w.value, k) == v
    assert w.value == m

def test_bidirectional():
    m = SomeModel()
    w = pn.panel(m, bidirectional=True)
    for k,v in alt_data.items():
        setattr(m, k, v)
        assert w._widgets[k].value == v
    assert w.value == m



import param
import pydantic

from typing import Dict, List, Any, Type, ClassVar, Optional

from pydantic import ValidationError, BaseModel
from pydantic.fields import FieldInfo, ModelField
from pydantic.config import inherit_config

from plum import dispatch, NotFoundLookupError

import param

import panel as pn
from copy import copy

from panel.layout import Column, Divider, ListPanel, Card

from panel.widgets import Widget, CompositeWidget, Button, LiteralInput

from .dispatchers import json_serializable, get_widget, clean_kwargs


class Config:
    """Pydantic Config overrides for monkey patching
    synchronization into a model.
    """

    validate_assignment = True


class pydantic_widgets(param.ParameterizedFunction):
    """Returns a dictionary of widgets to edit the fields
    of a pydantic model.
    """

    model = param.ClassSelector(pydantic.BaseModel, is_instance=False)

    aliases = param.Dict({})

    widget_kwargs = param.Dict({})
    defaults = param.Dict({})
    use_model_aliases = param.Boolean(False)
    callback = param.Callable()

    def __call__(self, **params):

        p = param.ParamOverrides(self, params)

        if isinstance(p.model, BaseModel):
            self.defaults = {f: getattr(p.model, f, None) for f in p.model.__fields__}
            
        if p.use_model_aliases:
            default_aliases = {
                field.name: field.alias.capitalize()
                for field in p.model.__fields__.values()
            }
        else:
            default_aliases = {
                name: name.replace("_", " ").capitalize() for name in p.model.__fields__
            }

        aliases = params.get("aliases", default_aliases)

        widgets = {}
        for field_name, alias in aliases.items():
            field = p.model.__fields__[field_name]

            value = p.defaults.get(field_name, None)

            if value is None:
                value = field.default

            value = json_serializable(value)

            try:
                widget_builder = get_widget.invoke(field.outer_type_, field.__class__)
                widget = widget_builder(value, field, name=field_name, **p.widget_kwargs)

            except (NotFoundLookupError, NotImplementedError):
                widget = get_widget(value, field, name=field_name, **p.widget_kwargs)

            if p.callback is not None:
                widget.param.watch(p.callback, "value")
                
            widgets[field_name] = widget
        return widgets


class InstanceOverride:
    """This allows us to override pydantic class attributes
    for specific instance without touching the instance __dict__
    since pydantic expects the instance __dict__ to only hold field
    values. We implement the descriptor protocol and lookup the value
    based on the id of the instance.
    """

    @classmethod
    def override(cls, instance: Any, name: str, value: Any, default: Any = None):
        """Override the class attribute `name` with `value`
        only when accessed from `instance`.

        Args:
            instance (Any): An instance of some class
            name (str): the attribute to be overriden
            value (Any): the value to override with for this instance
            default (Any, optional): Default value to return for other instances.
                                     Only used if attribute doesnt exist on class.
                                     Defaults to None.

        Returns:
            Any: the instance that was passed
        """

        class_ = type(instance)

        if not hasattr(class_, name):
            setattr(class_, name, cls(default))

        elif not isinstance(vars(class_)[name], cls):
            setattr(class_, name, cls(getattr(class_, name)))

        vars(class_)[name].mapper[id(instance)] = value

        return instance

    def revert_override(self, instance: Any):
        return self.mapper.pop(id(instance))

    def __init__(self, default, mapper=None):
        self.default = default
        self.mapper = mapper or {}

    def __get__(self, obj, objtype=None):
        if id(obj) in self.mapper:
            return self.mapper[id(obj)]
        else:
            return self.default


class PydanticModelEditor(CompositeWidget):
    """A composet widget whos value is a pydantic model and whos
    children widgets are synced with the model attributes

    """

    _composite_type: ClassVar[Type[ListPanel]] = Column
    _trigger_recreate: ClassVar[List] = ["class_"]
    
    _widgets = param.Dict(default={}, constant=True)

    _updating = param.Boolean(False)
    _updating_field = param.Boolean(False)

    class_ = param.ClassSelector(
        class_=BaseModel, default=None, is_instance=False
    )

    fields = param.List([])

    by_alias = param.Boolean(False)

    bidirectional = param.Boolean(False)

    value = param.ClassSelector(class_=(BaseModel, dict))

    def __init__(self, **params):
        
        super().__init__(**params)
        self._recreate_widgets()
        self.param.watch(self._recreate_widgets, 
                         self._trigger_recreate)

        self.param.watch(self._update_value, "value")

        if self.value is not None:
            self.param.trigger("value")

        for w in self.widgets:
            w.param.trigger("value")


    @property
    def widgets(self):
        fields = self.fields if self.fields else list(self._widgets)
        return [self._widgets[field] for field in fields]

    def _recreate_widgets(self, *events):
        if self.class_ is None:
            self.value = None
            return

        widgets = pydantic_widgets(
            model=self.class_,
            defaults=dict(self.items()),
            callback=self._validate_field,
            use_model_aliases=self.by_alias,
            widget_kwargs=dict(bidirectional=self.bidirectional),
        )

        with param.edit_constant(self):
            self._widgets = widgets


        self._composite[:] = self.widgets


    def _update_value(self, event: param.Event):
            
        if self._updating_field:
            return

        if self.value is None:
            for widget in self.widgets:
                try:
                    widget.value = None
                except:
                    pass
            return

        if self.class_ is None and isinstance(self.value, BaseModel):
            self.class_ = type(self.value)

        if isinstance(self.value, self.class_):
            for k,v in self.items():
                if k in self._widgets:
                    self._widgets[k].value = json_serializable(v)
                else:
                    self._recreate_widgets()
                    self.param.trigger('value')
                    return

        elif isinstance(self.value, dict) and not set(self.value).symmetric_difference(
            self._widgets
        ):
            self.value = self.class_(**self.value)
            return
        else:
            raise ValueError(
                f"value must be an instance of {self._class}"
                " or a dict matching its fields."
            )

        # HACK for biderectional sync
        if self.value is not None and self.bidirectional:

            # We need to ensure the model validates on assignment
            if not self.value.__config__.validate_assignment:
                config = inherit_config(Config, self.value.__config__)
                InstanceOverride.override(self.value, "__config__", config)

            # Add a callback to the root validators
            # to sync widgets to the changes made to
            # the model attributes
            callback = (False, self._update_widgets)
            if callback not in self.value.__post_root_validators__:
                validators = self.value.__post_root_validators__ + [callback]
                InstanceOverride.override(
                    self.value, "__post_root_validators__", validators
                )

            # If the previous value was a model
            # instance we unlink it by removing
            # the instance root validator and config
            if id(self.value) != id(event.old) and isinstance(event.old, BaseModel):
                for var in vars(type(event.old)).values():
                    if not isinstance(var, InstanceOverride):
                        continue
                    var.revert_override(event.old)

    def items(self):
        if self.value is None:
            return []
        return [(name, getattr(self.value, name)) 
                    for name in self.value.__fields__]

    def _validate_field(self, event: param.Event):
        if not event or self._updating:
            return

        if self.value is None:
            if self.class_ is not None:
                try:
                    data = {k: w.value for k, w in self._widgets.items()}
                    self.value = self.class_(**data)
                except:
                    pass
            return

        for name, widget in self._widgets.items():
            if event.obj == widget:
                break
        else:
            return

        field = self.value.__fields__[name]
        data = {k: w.value for k, w in self._widgets.items()}

        val = data.pop(name, None)
        val, error = field.validate(val, data, loc=name)
        if error:
            self.updating = True
            try:
                event.obj.value = event.old
            finally:
                self.updating = False
            raise ValidationError([error], type(self.value))

        if self.value is not None:
            setattr(self.value, name, val)
            self._updating_field = True
            try:
                self.param.trigger("value")
            finally:
                self._updating_field = False

    def _update_widgets(self, cls, values):
        if self.value is None:
            return

        if self._updating:
            return values

        self._updating = True
        try:
            for k, w in self._widgets.items():
                if k not in values:
                    continue
                val = json_serializable(values[k])
                if w.value != val:
                    w.value = val
        finally:
            self._updating = False

        return values

    @pn.depends("value")
    def json(self):
        if self.value is None:
            return pn.pane.JSON(width=self.width, sizing_mode="stretch_both")

        return pn.pane.JSON(
            object=self.value.json(), width=self.width, sizing_mode="stretch_both"
        )


class PydanticModelEditorCard(PydanticModelEditor):
    """Same as PydanticModelEditor but uses a Card container
    to hold the widgets and synces the header with the widget `name`
    """

    _composite_type: ClassVar[Type[ListPanel]] = Card
    collapsed = param.Boolean(False)

    def __init__(self, **params):
        super().__init__(**params)
        self._composite.header = self.name
        self.link(self._composite, name="header")
        self.link(self._composite, collapsed="collapsed")

import param

from pydantic_panel import get_widget

from typing import ClassVar, Type, List, Dict, Tuple, Any

import panel as pn
from panel.layout import Column, Divider, ListPanel, Card, Panel, Row

from panel.widgets import Widget, CompositeWidget, Button, LiteralInput


class BaseCollectionEditor(CompositeWidget):
    """Composite widget for editing a collections of items"""

    _composite_type: ClassVar[Type[ListPanel]] = Column
    
    _new_editor = param.Parameter()
    
    _widgets = param.Dict({})
    
    allow_add = param.Boolean(True)
    allow_remove = param.Boolean(True)
    
    item_added = param.Event()
    item_removed = param.Event()
    
    expand = param.Boolean(True)
    
    class_ = param.ClassSelector(object, is_instance=False)
    
    default_item = param.Parameter(default=None)
    
    value = param.Parameter(default=None)
    
    __abstract = True
    
    def __init__(self, **params):
        super().__init__(**params)
        self.param.watch(self._value_changed, 'value')
        self.param.trigger('value')
        
    def _panel_for(self, name, widget):
        panel = Card(widget, header=str(name), collapsed=not self.expand)
        
        if self.allow_remove:
            remove_button = Button(name='❌')
            def cb(event):
                self.remove_item(name)
            remove_button.on_click(cb)
            panel.append(remove_button)
        return panel
    
    def _create_widgets(self, *events, reset=True):
        if reset:
            self._widgets = {}
        for name, item in self.items():
            widget = self._widget_for(name, item)
            def cb(event):
                self.sync_item(name)
            widget.param.watch(cb, 'value')
            self._widgets[name] = widget
        
    def _update_panels(self, *events):
        panels = [self._panel_for(name, widget) 
                  for name, widget in self._widgets.items()]
        if self.name:
            panels.insert(0, pn.panel(f'## {self.name.capitalize()}'))
        panels.append(pn.panel(self._controls))
        panels.append(Divider())
        
        self._composite[:] = panels
        
    def _sync_widgets(self, *events):
        for name, item in self.items():
            widget = self._widgets.get(name, None)
            if widget is None:
                continue
            widget.value = item
    
    def _value_changed(self, *event):
        if not self.value:
            self._widgets = {}
            self._update_panels()
            return
        if set(self._widgets).symmetric_difference(self.keys()):
            self._create_widgets()
            self._update_panels()
        else:
            self._sync_widgets()
            
    def _controls(self):
        return pn.Column()
        
    def keys(self):
        raise NotImplementedError
        
    def values(self):
        raise NotImplementedError
        
    def items(self) -> List[Tuple[str,Any]]:
        raise NotImplementedError
    
    def add_item(self, item, name=None):
        raise NotImplementedError
    
    def remove_item(self, name):
        raise NotImplementedError
        
    def sync_item(self, name):
        raise NotImplementedError
        
    def _widget_for(self, name, item):
        raise NotImplementedError
        
    def _sync_values(self, *events):
        raise NotImplementedError
        

class ItemListEditor(BaseCollectionEditor):

    value = param.List(default=[])
    
    def keys(self):
        return list(range(len(self.value)))
        
    def values(self):
        return list(self.value)
        
    def items(self) -> List[Tuple[str,Any]]:
        return list(enumerate(self.value))
    
    def add_item(self, item, name=None):
        if name is None:
            name = len(self.value)
        idx = int(name)
        self.value.insert(idx, item)
        self.param.trigger('value')
        self.item_added = True
    
    def remove_item(self, name):
        self.value.pop(int(name))
        self.param.trigger('value')
        self.item_removed = True
    
    def sync_item(self, name):
        idx = int(name)
        self.value[idx] = self._widgets[idx].value
        self.param.trigger('value')
        
    def _add_new_cb(self, event):
        self.add_item(self.default_value)
    
    @param.depends('class_', 'allow_add')
    def _controls(self):
        if self.allow_add and self.class_ is not None:
            editor = self._widget_for(len(self.value), self.default_item)
            
            def cb(event):
                self.add_item(editor.value)
                
            add_button = Button(name='✅ Accept')
            add_button.on_click(cb)
            
            return Card(editor, add_button, header='➕ New', collapsed=True)
        
        return pn.Column()
    
    def _widget_for(self, name, item):
        if item is None:
            return get_widget.invoke(self.class_, None)(self.default_item, None, class_=self.class_, name=str(name))
        return get_widget(item, None, name=str(name))
    
    def _sync_values(self, *events):
        with param.parameterized.discard_events(self):
            self.value = [self._widgets[name].value for name in self.keys()]
        

class ItemDictEditor(BaseCollectionEditor):
    value = param.Dict(default={}, )
    
    key_type = param.ClassSelector(object, default=str, is_instance=False)
    
    default_key = param.Parameter(default='')
    
    def keys(self):
        return list(self.value)
        
    def values(self):
        return list(self.value.values())
        
    def items(self) -> List[Tuple[str,Any]]:
        return list(self.value.items())
    
    def add_item(self, item, name=None):
        if name is None:
            name = self.default_key
        self.value[name] = item
        self.param.trigger('value')
        self.item_added = True
                
    def remove_item(self, name):
        self.value.pop(name, None)
        self.param.trigger('value')
        self.item_removed = True
    
    def sync_item(self, name):
        self.value[name] = self._widgets[name].value
        self.param.trigger('value')
        
    def _widget_for(self, name, item):
        if item is None:
            return get_widget.invoke(self.class_, None)(item, None, class_=self.class_, name=str(name))
        return get_widget(item, None, name=str(name))
    
    def _sync_values(self, *events):
        with param.parameterized.discard_events(self):
            self.value = {name: self._widgets[name].value for name in self.keys()}
    

    @param.depends('class_', 'allow_add')
    def _controls(self):
        if self.allow_add and self.class_ is not None:
            key_editor = get_widget(self.default_key, None, name='Key', max_length=50)
            editor = self._widget_for(self.default_key, self.default_item)
            editor.name = 'Value'
            
            def cb(event):
                self.add_item(editor.value, key_editor.value)
                
            add_button = Button(name='✅ Accept')
            add_button.on_click(cb)
            
            return Card(key_editor, editor, add_button, header='➕ New', collapsed=True)
        return pn.Column()


@dispatch
def get_widget(value: BaseModel, field: Any, **kwargs):
    if field is None:
        class_= kwargs.pop('class_', type(value))
        return PydanticModelEditor(value=value, class_=class_, **kwargs)

    class_ = kwargs.pop('class_', field.outer_type_)
    kwargs = clean_kwargs(PydanticModelEditorCard, kwargs)
    return PydanticModelEditorCard(value=value, class_=class_, **kwargs)


@dispatch
def get_widget(value: List[BaseModel], field: Any, **kwargs):
    

    if field is not None:
        kwargs['class_'] = kwargs.pop('class_', field.type_)
        if value is None:
            value = field.default
    
    if value is None:
        value = []
    kwargs = clean_kwargs(ItemListEditor, kwargs)
    return ItemListEditor(value=value, **kwargs)

@dispatch
def get_widget(value: Dict[str,BaseModel], field: Any, **kwargs):
    
    if field is not None:
        kwargs['class_'] = kwargs.pop('class_', field.type_)
        if value is None:
            value = field.default
            
    if value is None:
        value = {}
 
    kwargs['key_type'] = kwargs.pop('key_type', str)
    kwargs = clean_kwargs(ItemDictEditor, kwargs)
    return ItemDictEditor(value=value, **kwargs)






"""Top-level package for pydantic-panel."""

__author__ = """Yossi Mosbacher"""
__email__ = "joe.mosbacher@gmail.com"
__version__ = "0.1.4"

from .dispatchers import json_serializable, get_widget
from .widgets import (
    PydanticModelEditor,
    PydanticModelEditorCard,
    ItemListEditor,
    ItemDictEditor,
)
from .pane import Pydantic



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pydantic_panel', 'tests']

package_data = \
{'': ['*']}

install_requires = \
['coverage>=6.4.2,<7.0.0',
 'panel>=0.12',
 'plum-dispatch',
 'pydantic',
 'pytest-cov>=3.0.0,<4.0.0']

setup_kwargs = {
    'name': 'pydantic-panel',
    'version': '0.1.4',
    'description': 'Top-level package for pydantic-panel.',
    'long_description': "==============\npydantic-panel\n==============\n\n\n.. image:: https://img.shields.io/pypi/v/pydantic_panel.svg\n        :target: https://pypi.python.org/pypi/pydantic_panel\n\n.. image:: https://img.shields.io/travis/jmosbacher/pydantic_panel.svg\n        :target: https://travis-ci.com/jmosbacher/pydantic_panel\n\n.. image:: https://readthedocs.org/projects/pydantic-panel/badge/?version=latest\n        :target: https://pydantic-panel.readthedocs.io/en/latest/?badge=latest\n        :alt: Documentation Status\n\n\nEdit pydantic models with panel.\n\nThis is just a small little project i made mostly for my own use and decided to share.\nIts limited in scope and probably still has bugs, USE AT YOUR OWN RISK.\n\nI will continue to add support for more types as I need them but feel free to \nopen issues with feature requests or better yet PRs with implementations.\n\n\n* Free software: MIT\n* Documentation: https://pydantic-panel.readthedocs.io.\n\nGetting Started\n---------------\n\nStep 1 - Install \n\n.. code-block::\n\n    pip install pydantic-panel\n\n\nStep 2 - Import pydantic_panel and add your models to layouts!\n\n.. code-block:: python\n    \n    import pydantic\n    import panel as pn\n    import pydantic_panel\n\n    class SomeModel(pydantic.BaseModel):\n        name: str\n        value: float\n\n    widget = pn.panel(SomeModel)\n\n    layout = pn.Column(widget, widget.json)\n    layout.servable()\n\n\nNow edit \n\nBasic Usage\n-----------\n\nIf you import `pydantic_panel`, it will register the widget automatically using the `panel.BasePane.applies` interface.\nAfter importing, calling `panel.panel(model)` will return a `panel.CompositeWidget` whos value is the model.\nWhen you change one of the sub-widget values, the new value is validated/coerced using the corresponding pydantic\nfield and if it passes validation/coercion the new value is set on the model itself.\nBy default this is a one-way sync, if the model field values are changed via code, it does not sync the widgets.\nIf you want biderectional sync, you can pass `bidirectional = True` to the widget constructor, this will patch the model \nto sync changes to the widgets but this may break without warning if pydantic change the internals of \ntheir `__setattr__` method.\n\n\n.. code-block:: python\n\n    import panel as pn\n    import pydantic_panel\n\n    class SomeModel(pydantic.BaseModel):\n        name: str\n        value: float\n\n    # when passing a model class, \n    # all widget values will be None including the composite widget value\n    w = pn.panel(SomeModel)\n    \n    # if you pass a model instance \n    # widget values will be the same as the model instance\n    inst = SomeModel(name='meaning', value=42)\n    w = pn.panel(inst)\n\n    # This will display widgets to e.g. edit the model in a notebook\n    w\n\n    # This will return True\n    inst is w.value\n\n    # This will be None if the widgets have not yet been set to values\n    # if all the required fields have been set, this will be an instance of SomeModel\n    # with the validated attribute values from the widgets\n    w.value\n\n\nThe `pn.panel` method will return a widget which can be used as part of a larger application or as just \na user friendly way to edit your model data in the notebook.\n\nCustomizing widgets\n-------------------\n\nYou can add or change the widgets used for a given type by hooking into the dispatch\nmechanism (we use plum-dispatch). This can be used to override the widget used for a supported\ntype or to add supprt for a new type.\n\n\n.. code-block:: python\n\n    from pydantic_panel import get_widget\n    from pydantic import FieldInfo\n\n    # precedence = 1 will ensure this function will be called\n    # instead of the default which has precedence = 0\n    @get_widget.dispatch(precedence=1)\n    def get_widget(value: MY_TYPE, field: FieldInfo, **kwargs):\n        # extract relavent info from the pydantic field info here.\n\n        # return your favorite widget\n        return MY_FAVORITE_WIDGET(value=value, **kwargs)\n\n\nSupporting non-serializable types\n---------------------------------\n\nPanel encodes the data sent to the widget using json serialization. \nIf your type is not json serializable, you can have pydantic-panel convert\nthe data to a json-serializable object before its passed to the widget. To add\nthis conversion, register a conversion function using the `json_serializable.dispatch`\ndecorator\n\n.. code-block:: python\n\n    from pydantic_panel import json_serializable\n\n    # precedence = 1 will ensure this function will be called\n    # instead of the default which has precedence = 0\n    @json_serializable.dispatch(precedence=1)\n    def json_serializable(value: TYPE):\n        # convert to a serializable object\n        value = some_function(value)\n        return value\n\n\nSupported types\n---------------\n\n* int\n* float\n* str\n* list\n* tuple\n* dict\n* datetime.datetime\n* BaseModel\n* List[BaseModel]\n* pandas.Interval\n* numpy.ndarray\n\nFAQ\n---\n\nQ: Why did you decide to use CompositWidget instead of Pane like Param uses?\n\nA: Nested models. This is a recursive problem, so I was looking for a recursive solution. By using a Widget to\ndisplay models, all fields are treated equally. A field of type BaseModel is edited with a widget that has a `.value` \nattribute just like any other field and therefore requires no special treatment. When the parent collects the values of its children \nit just reads the `widget.value` attribute and does not need to check whether the value is nested or not. At every level \nof the recursion the widget only has to care about the fields on its model class and watch only the `.value` attribute of\nits children widgets for changes.\n\n\nFeatures\n--------\n\n* TODO\n\nCredits\n-------\n\nThis package was created with Cookiecutter_ and the `briggySmalls/cookiecutter-pypackage`_ project template.\n\n.. _Cookiecutter: https://github.com/audreyr/cookiecutter\n.. _`briggySmalls/cookiecutter-pypackage`: https://github.com/briggySmalls/cookiecutter-pypackage\n",
    'author': 'Yossi Mosbacher',
    'author_email': 'joe.mosbacher@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/jmosbacher/pydantic_panel',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4',
}


setup(**setup_kwargs)
